<?php
/**
 * Plugin Name:         RS Bibliography
 * Description:         A really simple bibliography system with references stored in a taxonomy.
 * Plugin URI:          https://plugins.seindal.dk/plugins/rs-bibliography/
 * Update URI:          https://plugins.seindal.dk
 * Author:              René Seindal
 * Author URI:          https://plugins.seindal.dk/
 * Donate link:         https://mypos.com/@historywalks
 * Requires Plugins:    rs-base-plugin
 * License:             GPL v2 or later
 * License URI:         https://www.gnu.org/licenses/gpl-2.0.html
 * Text Domain:         rs-bibliography
 * Domain Path:         /languages
 * Requires PHP:        7.4
 * Requires at least:   5.0
 * Version:             0.5.2
 **/

namespace ReneSeindal;

if ( ! defined( 'ABSPATH' ) ) exit; // Exit if accessed directly

add_action( 'plugins_loaded', function() {
    class Bibliography extends PluginBase {
        protected $plugin_file = __FILE__;

        use PluginBaseSettings;

        protected $taxonomy = 'bibliography';

        /*************************************************************
         *
         * Registers the `bibliography` taxonomy
         *
         *************************************************************/

        function do_init_action_register_bibliography_taxonomy() {
            register_taxonomy( $this->taxonomy, [ 'post' ], [
                'hierarchical'          => false,
                'public'                => true,
                'show_in_nav_menus'     => true,
                'show_ui'               => true,
                'show_admin_column'     => true,
                'query_var'             => true,
                'rewrite'               => true,
                'capabilities'          => [
                    'manage_terms' => 'edit_posts',
                    'edit_terms'   => 'edit_posts',
                    'delete_terms' => 'edit_posts',
                    'assign_terms' => 'edit_posts',
                ],
                'labels'                => [
                    'name'                       => __( 'References', 'rs-bibliography' ),
                    'singular_name'              => _x( 'Reference', 'taxonomy general name', 'rs-bibliography' ),
                    'search_items'               => __( 'Search References', 'rs-bibliography' ),
                    'popular_items'              => __( 'Popular References', 'rs-bibliography' ),
                    'all_items'                  => __( 'All References', 'rs-bibliography' ),
                    'parent_item'                => __( 'Parent Reference', 'rs-bibliography' ),
                    'parent_item_colon'          => __( 'Parent Reference:', 'rs-bibliography' ),
                    'edit_item'                  => __( 'Edit Reference', 'rs-bibliography' ),
                    'update_item'                => __( 'Update Reference', 'rs-bibliography' ),
                    'view_item'                  => __( 'View Reference', 'rs-bibliography' ),
                    'add_new_item'               => __( 'Add New Reference', 'rs-bibliography' ),
                    'new_item_name'              => __( 'New Reference', 'rs-bibliography' ),
                    'separate_items_with_commas' => __( 'Separate references with commas', 'rs-bibliography' ),
                    'add_or_remove_items'        => __( 'Add or remove references', 'rs-bibliography' ),
                    'choose_from_most_used'      => __( 'Choose from the most used references', 'rs-bibliography' ),
                    'not_found'                  => __( 'No references found.', 'rs-bibliography' ),
                    'no_terms'                   => __( 'No references', 'rs-bibliography' ),
                    'menu_name'                  => __( 'References', 'rs-bibliography' ),
                    'items_list_navigation'      => __( 'References list navigation', 'rs-bibliography' ),
                    'items_list'                 => __( 'References list', 'rs-bibliography' ),
                    'most_used'                  => _x( 'Most Used', 'reference', 'rs-bibliography' ),
                    'back_to_items'              => __( '&larr; Back to References', 'rs-bibliography' ),
                ],
                'show_in_rest'          => true,
                'rest_base'             => $this->taxonomy,
                'rest_controller_class' => 'WP_REST_Terms_Controller',

                'update_count_callback' => '_update_generic_term_count',
            ] );

        }

        /*
         *
         * Sets the post updated messages for the `reference` taxonomy.
         *
         * @param  array $messages Post updated messages.
         * @return array Messages for the `reference` taxonomy.
         *
         */

        function do_term_updated_messages_filter_bibliography_updated_messages( $messages ) {

            $messages[$this->taxonomy] = [
                0 => '', // Unused. Messages start at index 1.
                1 => __( 'Reference added.', 'rs-bibliography' ),
                2 => __( 'Reference deleted.', 'rs-bibliography' ),
                3 => __( 'Reference updated.', 'rs-bibliography' ),
                4 => __( 'Reference not added.', 'rs-bibliography' ),
                5 => __( 'Reference not updated.', 'rs-bibliography' ),
                6 => __( 'References deleted.', 'rs-bibliography' ),
            ];

            return $messages;
        }


        /************************************************************
         *
         * Add filter on post lists on this taxonomy
         *
         ************************************************************/

        function do_restrict_manage_posts_action( $post_type, $which ) {
            if ( ! $this->get_option( 'post_list_filter', false ) )
                return;

            $tax = get_taxonomy( $this->taxonomy );

            if ( in_array( $post_type, $tax->object_type ) ) {

                $selected = '';
                if ( isset( $_REQUEST[ $this->taxonomy ] ) ) {
                    $selected = $_REQUEST[ $this->taxonomy ];
                }

                wp_dropdown_categories(array(
                    'show_option_all' =>  $tax->labels->all_items,
                    'taxonomy'        =>  $this->taxonomy,
                    'value_field'     => 'slug',
                    'name'            =>  $this->taxonomy,
                    'orderby'         =>  'name',
                    'selected'        =>  $selected,
                    'hierarchical'    =>  false,
                    'show_count'      =>  true,
                    'hide_empty'      =>  false,
                ) );
            }
        }


        /************************************************************
         *
         * Add columms to bibliography table
         *
         ************************************************************/

        protected $manage_edit_bibliography_columns_callback_hook = 'manage_edit-bibliography_columns';

        function do_manage_edit_bibliography_columns_filter( $columns ) {
            $columns['bib_short'] = __( 'Short ref', 'rs-bibliography' );
            $columns['bib_full'] = __( 'Full ref', 'rs-bibliography' );
            return $columns;
        }

        function do_manage_bibliography_custom_column_action( $string, $column, $term_id ) {
            switch ( $column ) {
            case 'bib_short':
                echo $this->format_reference_short( $term_id );
                break;
            case 'bib_full':
                echo $this->format_reference_long( $term_id );
                break;
            }
        }


        /************************************************************
         *
         * Custom fields for bibliography entries
         *
         ************************************************************/

        protected function bibliography_fields() {
            return [
                'bib_author' => [
                    'name' => __( 'Author', 'rs-bibliography' ),
                    'description' => __( 'Name of the author(s) -- separate multiple authors with //.', 'rs-bibliography' ),
                ],
                'bib_title' => [
                    'name' => __( 'Title', 'rs-bibliography' ),
                    'description' => __( 'The title of the work.', 'rs-bibliography' ),
                ],
                'bib_serial' => [
                    'name' => __( 'Serial / Collection', 'rs-bibliography' ),
                    'description' => __( 'The series, journal or collection the work is in.', 'rs-bibliography' ),
                ],
                'bib_editor' => [
                    'name' => __( 'Publisher', 'rs-bibliography' ),
                    'description' => __( 'The publisher and/or editor of the work.', 'rs-bibliography' ),
                ],
                'bib_year' => [
                    'name' => __( 'Year', 'rs-bibliography' ),
                    'description' => __( 'The publishing year of the work.', 'rs-bibliography' ),
                ],
            ];
        }

        function do_bibliography_add_form_fields_action() {
            $fields = $this->bibliography_fields();

            foreach ( $fields as $field => $labels ) {
                echo '<div class="form-field">';
                printf( '<label for="%s">%s</label>',
                        esc_attr( $field ), esc_html( $labels['name'] ) );
                printf( '<input name="%s" id="%s" type="text" value="">',
                        esc_attr( $field ), esc_attr( $field ) );
                printf( '<p id="%s-description">%s</p>',
                        esc_attr( $field ), esc_html( $labels['description'] ) );
                echo '</div>';
            }
        }

        function do_bibliography_edit_form_fields_action( $term ) {
            $fields = $this->bibliography_fields();

            foreach ( $fields as $field => $labels ) {
                $value = trim( get_term_meta( $term->term_id, $field, true ) );

                echo '<tr class="form-field">';
                printf( '<th scope="row"><label for="%s">%s</label></th>',
                        esc_attr( $field ), esc_html( $labels['name'] ));
                printf( '<td><input name="%s" id="%s" type="text" value="%s">',
                        esc_attr( $field ), esc_attr( $field ), esc_attr( $value ));
                printf( '<p class="description">%s</p></td>',
                        esc_html( $labels['description'] ));
                echo '</tr>';
            }
        }

        protected function store_term_field( $term_id, $query_var, $field = NULL ) {
            if ( !isset( $_REQUEST[$query_var] ) )
                return;

            $value = trim( $_REQUEST[$query_var] ?? '' );

            if ( !$field )
                $field = $query_var;

            if ( empty( $value ) )
                delete_term_meta( $term_id, $field );
            else
                update_term_meta( $term_id, $field, $value );
        }

        function do_create_bibliography_action( $term_id ) {
            $fields = $this->bibliography_fields();
            foreach ( $fields as $field => $labels )
                $this->store_term_field( $term_id, $field );
        }

        function do_edited_bibliography_action( $term_id ) {
            $this->do_create_bibliography_action( $term_id );
        }


        /************************************************************
         *
         * Bibliography entry formatting -- site dependent, but can be
         * overridden by filters.
         *
         * Defaults are:
         * Long form: AUTHOR. _TITLE_. EDITOR, YEAR.
         * Short form: AUTHOR (YEAR)
         *
         ************************************************************/

        protected function get_term_object( $term ) {
            $fields = $this->bibliography_fields();

            $data= [];
            foreach ( array_keys( $fields ) as $field ) {
                $pretty = $field;
                if ( str_starts_with( $field, 'bib_' ) )
                    $pretty = substr( $field, 4 );

                $data[$pretty] = trim( get_term_meta( $term->term_id, $field, true ) );
            }

            if ( !empty( $data['author'] ) ) {
                $authors = $this->parse_authors( $data['author'] );

                if ( count( $authors ) > 1 && str_starts_with( $authors[count( $authors )-1][0], '(' ) )
                    $data['eds'] = ( array_pop( $authors ) )[0];

                $data['authors'] = $authors;
            }

            return (object)$data;
        }

        // Parse author lists separated by // with optional (eds.) at the end
        public function parse_authors( $author ) {
            return array_map(
                fn( $a ) => array_map(
                    fn ( $a ) => trim( $a ),
                    explode( ',', $a, 2 )
                ),
                explode( '//', $author )
            );
        }

        // Generate short reference
        protected function format_reference_short( $term ) {
            $term = get_term( $term );
            return apply_filters( 'rs_bibliography_bibrefs', '', $this->get_term_object( $term ), $term );
        }

        // Default short reference
        function do_rs_bibliography_bibrefs_filter( $text, $ref, $term ) {
            if ( $text )
                return $text;

            if ( empty( $ref->author ) ) {
                if ( empty( $ref->title ) )
                    return esc_html( $term->name ); // fallback no author no title

                list ($first,$last) = preg_split( '/[,.;:()-]/', $ref->title . '.', 2 );
                $author =  sprintf( '<i>%s%s</i>', esc_html( $first ), ( empty( $last ) ? '' : '...' ) );
            } else {
                if ( count( $ref->authors ) > 1 )
                    $author = sprintf( '%s <i>et.al.</i>', esc_html( $ref->authors[0][0] ) );
                else
                    $author = esc_html( $ref->authors[0][0] );
            }

            return sprintf( '%s (%04d)', $author, $ref->year ?? __( 'n.d.', 'rs-bibliography' ) );
        }

        // Generate log reference
        protected function format_reference_long( $term ) {
            $term = get_term( $term );
            return apply_filters( 'rs_bibliography_bibrefl', '', $this->get_term_object( $term ), $term );
        }

        function do_rs_bibliography_bibrefl_filter( $text, $ref, $term ) {
            if ( $text )
                return $text;

            $periods = [];

            if ( !empty( $ref->author ) )
                $periods[] = esc_html( $this->format_author_full( $ref->authors, $ref->eds ?? NULL ) );

            if ( !empty( $ref->title ) ) {
                if ( empty( $ref->serial ) ) {
                    if ( str_contains( $ref->title, '//' ) ) {
                        list ( $title, $material ) = array_map( fn( $s ) => trim( $s ),  explode( '//', $ref->title ));
                        $periods[] = sprintf( '<i>%s</i>, %s', esc_html( $title ), esc_html( $material ) );
                    }  else
                        $periods[] = sprintf( '<i>%s</i>', esc_html( $ref->title ) );
                } else
                    $periods[] = sprintf( '<i>%s</i> in %s', esc_html( $ref->title ), esc_html( $ref->serial ) );
            }

            $wherewhen = [];
            if ( !empty( $ref->editor ) )
                $wherewhen[] = esc_html( $ref->editor );
            if ( !empty( $ref->year ) )
                $wherewhen[] = esc_html( $ref->year );

            if ( !empty( $wherewhen ) )
                $periods[] = join(', ', array_map( fn( $s ) => rtrim( $s, ',' ), $wherewhen ) );

            if ( empty( $periods ) )
                $periods[] = esc_html( $term->name );


            return join( '. ', array_map( fn( $s ) => rtrim( $s, '.' ), $periods ) );
        }

        // Format a full list of authors or editors
        public function format_author_full( $authors, $eds ) {
            // List of formatted names
            $names = [];

            // First author is always "LAST, FIRST"
            $first = array_shift( $authors );
            $names[] = join(', ', $first );

            // Subsequent authors are "FIRST LAST"
            foreach ( $authors as $a )
                $names[] = join(' ', array_reverse( $a ) );

            // Separate authors by commas, but 'and' before the last
            $last = array_pop( $names );
            if ( empty( $names ) )
                $namelist = $last;
            else {
                $namelist = join( ', ', $names ) . ' and ' . $last;
            }

            // Don't for editors
            if ( $eds )
                return $namelist . ' ' . $eds;

            return $namelist;
        }

        /************************************************************
         *
         * Bibliography shortcodes
         *
         ************************************************************/

        // Shortcode [bibliography] outputs a bibliography based on tags used.

        function do_bibliography_shortcode( $atts = [], $content = null, $tag = '' ) {
            // normalize attribute keys, lowercase
            $atts = array_change_key_case( (array)$atts, CASE_LOWER );

            // override default attributes with user attributes
            $atts = shortcode_atts( [
                'header' => false,
                'html' => false,
            ] , $atts, $tag );

            $post = get_post();

            if ( empty( $post ) )
                return '';

            $terms = get_the_terms( $post, $this->taxonomy );
            if ( !is_array( $terms ) )
                return '';

            usort( $terms, fn($a,$b) => ( strtolower( $a->name ) <=> strtolower( $b->name ) ) );

            $output = [];

            if ( false === $atts['header'] )
                $atts['header'] = $this->get_option( 'title' );

            if ( false === $atts['html'] )
                $atts['html'] = $this->get_option( 'title_tag' );

            if ( !empty( $atts['header'] ) && !empty( $atts['html'] ) )
                $output[] = sprintf( '<%s class="bibliography full-bibliography">%s</%s>',
                                     esc_attr( $atts['html'] ),
                                     esc_html( $atts['header'] ),
                                     esc_attr( $atts['html'] )
                );

            foreach ( $terms as $term )
                $output[] = sprintf( '<p class="bibliography full-bibliography">%s.</p>', $this->format_reference_long( $term ) );

            return join( "\n", $output );
        }

        // Shortcode [full_bibliography] outputs everything in the
        // taxonomy with a non-zero count

        function do_full_bibliography_shortcode( $atts = [], $content = null, $tag = '' ) {
            // normalize attribute keys, lowercase
            $atts = array_change_key_case( (array)$atts, CASE_LOWER );

            // override default attributes with user attributes
            $atts = shortcode_atts( [
                'header' => false,
                'html' => false,
            ] , $atts, $tag );

            $terms = get_terms( [ 'taxonomy' => $this->taxonomy] );
            if ( !is_array( $terms ) )
                return '';

            usort( $terms, fn($a,$b) => ( strtolower( $a->name ) <=> strtolower( $b->name ) ) );

            $output = [];

            if ( false === $atts['header'] )
                $atts['header'] = $this->get_option( 'fulltitle' );

            if ( false === $atts['html'] )
                $atts['html'] = $this->get_option( 'fulltitle_tag' );

            if ( !empty( $atts['header'] ) && !empty( $atts['html'] ) )
                $output[] = sprintf( '<%s class="bibliography">%s</%s>',
                                     esc_attr( $atts['html'] ),
                                     esc_html( $atts['header'] ),
                                     esc_attr( $atts['html'] )
                );

            foreach ( $terms as $term )
                $output[] = sprintf( '<p class="bibliography">%s.</p>', $this->format_reference_long( $term ) );

            return join( "\n", $output );
        }

        // Shortcodes [bibref* slug=XXX] or just [bibref* XXX] outputs a
        // reference to a work, whether tagged or not.

        function bibref_shortcode_wrapper( $atts = [], $content = null, $tag = '' ) {
            // normalize attribute keys, lowercase
            $atts = array_change_key_case( (array)$atts, CASE_LOWER );

            // override default attributes with user attributes
            $atts = shortcode_atts( [
                0 => NULL,          // XXX Hrmff
                'slug' => NULL,
                'link' => false,
            ] , $atts, $tag );

            $slug = $atts['slug'] ?? $atts[0];

            if ( $slug )
                $term = get_term_by( 'slug', $slug, $this->taxonomy );

            if ( ! is_a( $term, 'WP_Term' ) )
                return sprintf( __( '[[BIBREF ERROR slug=%s]]', 'rs-bibliography' ), esc_html( $slug ) );

            $output = apply_filters( '_rs_bibliography_' . $tag, $term->name, $slug, $term );

            if ( $atts['link'] )
                $output = sprintf( '<a href="%s">%s</a>', get_term_link( $term ), $output );

            return $output;
        }

        protected $last_ref = NULL;
        protected $seen_refs = [];

        function do_bibrefs_shortcode( $atts = [], $content = null, $tag = '' ) {
            return $this->bibref_shortcode_wrapper( $atts, $content, $tag );
        }

        function do__rs_bibliography_bibrefl_filter( $output, $slug, $term ) {
            return $this->format_reference_long( $term );
        }

        function do_bibrefl_shortcode( $atts = [], $content = null, $tag = '' ) {
            return $this->bibref_shortcode_wrapper( $atts, $content, $tag );
        }

        function do__rs_bibliography_bibrefs_filter( $output, $slug, $term ) {
            return $this->format_reference_short( $term );
        }

        function do_bibref_shortcode( $atts = [], $content = null, $tag = '' ) {
            return $this->bibref_shortcode_wrapper( $atts, $content, $tag );
        }

        function do__rs_bibliography_bibref_filter( $output, $slug, $term ) {
            if ( $this->get_option( 'bibref_ibid' ) )
                if ( isset( $this->last_ref ) && $this->last_ref == $slug )
                    return __( '<i>ibid.</i>', 'rs-bibliography' );

            $this->last_ref = $slug;

            if ( $this->get_option( 'bibref_short_only' ) || array_key_exists( $slug, $this->seen_refs ) )
                return $this->format_reference_short( $term );

            $this->seen_refs[$slug] = true;

            return $this->format_reference_long( $term );
        }

        /*************************************************************
         *
         * Connect taxonomy and post types
         *
         *************************************************************/

        function do_wp_loaded_action() {
            $taxonomy = get_taxonomy( $this->taxonomy );
            $post_types = $this->get_option( 'post_types', $taxonomy->object_type );

            foreach ( $post_types as $post_type )
                if ( post_type_exists( $post_type ) )
                    register_taxonomy_for_object_type( $this->taxonomy, $post_type );
        }



        /************************************************************************
         *
         *	Settings
         *
         ************************************************************************/

        function do_admin_menu_action() {
            $this->settings_add_menu(
                __( 'Bibliography', 'rs-bibliography'),
                __( 'RS Bibliography', 'rs-bibliography' )
            );
        }

        // Called automatically by PluginBaseSettings
        function settings_define_sections_and_fields() {
            $section = $this->settings_add_section(
                'section_post_types',
                __( 'Post-types', 'rs-bibliography' ),
                __( 'Select the post-types you want bibliographies in.', 'rs-bibliography' )
            );

            $taxonomy = get_taxonomy( $this->taxonomy );

            $this->settings_build_post_types_menu(
                'post_types', $section,
                __( 'Post-types', 'rs-bibliography' ),
                [ 'public' => true ],
                $taxonomy->object_type,
                'rs_bibliography_post_types'
            );

            $section = $this->settings_add_section(
                'section_integration',
                __( 'Integrations', 'rs-bibliography' ),
                __( 'Select how this plugin integrates into the WordPress admin interface', 'rs-bibliography' )
            );

            $this->settings_add_field(
                'post_list_filter', $section,
                __( 'Allow post filtering', 'rs-bibliography' ),
                'settings_field_checkbox_html'
            );


            $headers = [
                '' => __( 'No automatic header', 'rs-bibliography' ),
                'h1' => __( 'Level 1 (<H1>)', 'rs-bibliography' ),
                'h2' => __( 'Level 2 (<H2>)', 'rs-bibliography' ),
                'h3' => __( 'Level 3 (<H3>)', 'rs-bibliography' ),
                'h4' => __( 'Level 4 (<H4>)', 'rs-bibliography' ),
                'h5' => __( 'Level 5 (<H5>)', 'rs-bibliography' ),
                'h6' => __( 'Level 6 (<H6>)', 'rs-bibliography' ),
                'div' => __( '<DIV class="bibliography">', 'rs-bibliography' ),
            ];

            $section = $this->settings_add_section(
                'section_post_bibliography',
                __( 'Single post bibliography', 'rs-bibliography' ),
                __( 'Heading output by [bibliography] shortcode.', 'rs-bibliography' )
            );

            $this->settings_add_field(
                'title', $section,
                __( 'Heading', 'rs-bibliography' ),
                'settings_field_input_html',
                [
                    'default' => __( 'Bibliography', 'rs-bibliography' ),
                    'size' => 40,
                ]
            );

            $this->settings_add_field(
                'title_tag', $section,
                __( 'HTML wrapper', 'rs-bibliography' ),
                'settings_field_select_html',
                [
                    'values' => $headers,
                    'default' => 'h2',
                ]
            );

            $section = $this->settings_add_section(
                'section_full_bibliography',
                __( 'Full bibliography', 'rs-bibliography' ),
                __( 'Heading output by [full_bibliography] shortcode.', 'rs-bibliography' )
            );

            $this->settings_add_field(
                'fulltitle', $section,
                __( 'Heading', 'rs-bibliography' ),
                'settings_field_input_html',
                [
                    'default' => __( 'Site bibliography', 'rs-bibliography' ),
                    'size' => 40,
                ]
            );

            $this->settings_add_field(
                'fulltitle_tag', $section,
                __( 'HTML wrapper', 'rs-bibliography' ),
                'settings_field_select_html',
                [
                    'values' => $headers,
                    'default' => 'h2',
                ]
            );

            $section = $this->settings_add_section(
                'section_bibref',
                __( 'The [bibref] shortcode', 'rs-bibliography' ),
                [
                    __( 'Details of the outpug of the [bibref] shortcode.', 'rs-bibliography' ),
                    __( 'Default behaviour is to use the long form the first time a work is referenced, then the short form, but for consecutive references to the same work, replace the short form with "ibid."', 'rs-bibliography' ),
                ]
            );

            $this->settings_add_field(
                'bibref_short_only', $section,
                __( 'Always use the short form', 'rs-bibliography' ),
                'settings_field_checkbox_html'
            );

            $this->settings_add_field(
                'bibref_ibid', $section,
                __( 'Use "ibid."', 'rs-bibliography' ),
                'settings_field_checkbox_html'
            );
        }
    }

    Bibliography::instance();
} );
